from version_differ.common import *
import requests
import json
import os
import sys
from zipfile import ZipFile
import tarfile
from os.path import join
import shutil


def download_package_source(url, ecosystem, package, version, dir_path):
    print("fetching {}-{} in {} ecosystem from {}".format(package, version, ecosystem, url))

    # First try based on file extension
    if url.endswith(".whl") or url.endswith(".jar") or url.endswith(".zip"):
        download_zipped(url, dir_path)
    elif url.endswith(".gz") or url.endswith(".crate") or url.endswith(".gem") or url.endswith(".tgz"):
        download_tar(url, dir_path)
    elif ecosystem == COMPOSER or ecosystem == MAVEN:
        download_zipped(url, dir_path)
    elif ecosystem == NPM or ecosystem == PYPI or ecosystem == RUBYGEMS or ecosystem == CARGO:
        download_tar(url, dir_path)
    else:
        # do nothing
        return None

    path = None
    if ecosystem == COMPOSER or ecosystem == NPM or ecosystem == CARGO:
        files = os.listdir(dir_path)
        assert len(files) == 1
        path = "{}/{}".format(dir_path, files[0])
    elif ecosystem == PIP:
        files = os.listdir(dir_path)
        if len(files) == 1:
            # for tar.gz extractions
            path = "{}/{}".format(dir_path, files[0])
        else:
            # assuming wheel file
            distinfo = None
            for f in files:
                if f.endswith(".dist-info"):
                    distinfo = f
                    break
            if distinfo:
                shutil.rmtree(join(dir_path, distinfo), ignore_errors=True)
                path = dir_path
    elif ecosystem == MAVEN or ecosystem == RUBYGEMS:
        path = dir_path
    else:
        files = os.listdir(dir_path)
        sys.exit("check downloding regstiry tarball for:{}-{}".format(ecosystem, files))
    assert path, "cannot extract {}-{}".format(ecosystem, package)
    return path


def get_package_version_source_url(ecosystem, package, version):
    assert ecosystem in ecosystems

    if ecosystem == CARGO:
        return "https://crates.io/api/v1/crates/{}/{}/download".format(package, version)
    elif ecosystem == COMPOSER:
        url = "https://repo.packagist.org/packages/{}.json".format(package)
        data = json.loads(requests.get(url).content)["package"]["versions"]
        data = {k[1:] if k.startswith("v") else k: v for k, v in data.items()}
        if version in data:
            return data[version]["dist"]["url"]
    elif ecosystem == NPM:
        url = "https://registry.npmjs.org/{}".format(package)
        data = json.loads(requests.get(url).content)["versions"]
        data = {k[1:] if k.startswith("v") else k: v for k, v in data.items()}
        if version in data:
            return data[version]["dist"]["tarball"]
    elif ecosystem == PYPI:
        url = "https://pypi.org/pypi/{}/json".format(package)
        data = json.loads(requests.get(url).content)["releases"]
        data = {k[1:] if k.startswith("v") else k: v for k, v in data.items()}
        if version in data:
            data = data[version]
            url = next((x["url"] for x in data if x["url"].endswith(".whl")), data[-1]["url"])
            return url
    elif ecosystem == RUBYGEMS:
        return "https://rubygems.org/downloads/{}-{}.gem".format(package, version)
    elif ecosystem == MAVEN:
        url = get_maven_pacakge_url(package)
        if url:
            artifact = package.split(":")[1]
            url = "{}/{}/{}-{}-sources.jar".format(url, version, artifact, version)
            if requests.get(url).status_code == 200:
                return url

    return None


def download_zipped(url, path):
    compressed_file_name = "temp_data.zip"
    dest_file = join(path, compressed_file_name)

    r = requests.get(url, stream=True)
    with open(dest_file, "wb") as output_file:
        output_file.write(r.content)

    z = ZipFile(dest_file, "r")
    z.extractall(path)
    z.close()
    os.remove(dest_file)


def download_tar(url, path):
    compressed_file_name = "temp_data.tar.gz"
    dest_file = join(path, compressed_file_name)

    r = requests.get(url)
    with open(dest_file, "wb") as output_file:
        output_file.write(r.content)

    t = tarfile.open(dest_file)
    t.extractall(path)
    t.close()
    os.remove(dest_file)

    # additional logic for ruby gems
    if url.endswith(".gem"):
        ruby_tar = "data.tar.gz"
        assert ruby_tar in os.listdir(path)
        ruby_tar = join(path, ruby_tar)

        t = tarfile.open(ruby_tar)
        t.extractall(path)
        t.close()
        os.remove(ruby_tar)

        for gz_file in ["metadata.gz", "checksums.yaml.gz"]:
            if gz_file in os.listdir(path):
                os.remove(join(path, gz_file))


def get_maven_pacakge_url(package):
    url = "https://repo1.maven.org/maven2/" + package.replace(".", "/").replace(":", "/")
    if requests.get(url).status_code == 200:
        return url

    s1, s2 = package.split(":")
    url = "https://repo1.maven.org/maven2/" + s1.replace(".", "/") + "/" + s2
    if requests.get(url).status_code == 200:
        return url



"""Main module."""

import json
from pygit2 import init_repository, Signature
from time import time
import tempfile
from git import Repo
import re
from unidiff import PatchSet
from urllib.parse import urlparse, parse_qs
from os.path import join, relpath
import os
from version_differ.download import download_package_source, get_package_version_source_url
from version_differ.common import *


class VersionDifferOutput:
    def __init__(self):
        self.old_version = None
        self.old_version_git_sha = None
        self.new_version = None
        self.new_version_git_sha = None

        self.diff = None

        self.new_version_filelist = None
        self.old_version_filelist = None

    def to_json(self):
        return {
            "metadata_info": {
                "old_version": self.old_version,
                "old_version_git_sha": self.old_version_git_sha,
                "new_version": self.new_version,
                "new_version_git_sha": self.new_version_git_sha,
            },
            "diff": self.diff,
            "new_version_filelist": self.new_version_filelist,
            "old_version_filelist": self.old_version_filelist,
        }


class FileDiff:
    def __init__(self, source_file, target_file, is_rename, loc_added, loc_removed, added_lines, removed_lines):
        self.source_file = source_file
        self.target_file = target_file
        self.is_rename = is_rename
        self.loc_added = loc_added
        self.loc_removed = loc_removed
        self.added_lines = added_lines
        self.removed_lines = removed_lines

    # TODO: convert to json for cli output


def sanitize_repo_url(repo_url):
    parsed_url = urlparse(repo_url)
    host = parsed_url.netloc

    if host == "gitbox.apache.org" and "p" in parse_qs(parsed_url.query).keys():
        project_name = parse_qs(parsed_url.query)["p"]
        return "https://gitbox.apache.org/repos/asf/{}".format(project_name)

    if host == "svn.opensymphony.com":
        return repo_url

    # below rule covers github, gitlab, bitbucket, foocode, eday, qt
    sources = ["github", "gitlab", "bitbucket", "foocode", "eday", "q", "opendev"]
    assert any([x in host for x in sources]), "unknown host for repository url: {}".format(repo_url)

    paths = [s.removesuffix(".git") for s in parsed_url.path.split("/")]
    owner, repo = paths[1], paths[2]
    return "https://{}/{}/{}".format(host, owner, repo)


def get_commit_of_release(tags, package, version):
    """
    tags: gitpython object,
    version: string, taken from ecosystem data
    """

    # Now we check through a series of heuristics if tag matches a version
    version_formatted_for_regex = version.strip().replace(".", "\\.")
    package = package.lower()
    patterns = [
        # 1. Ensure the version part does not follow any digit between 1-9,
        # e.g., to distinguish betn 0.1.8 vs 10.1.8
        r"^(?:.*[^1-9])?{}$".format(version_formatted_for_regex),
        # 2. check if and only if crate name and version string is present
        # besides non-alphanumeric, e.g., to distinguish guppy vs guppy-summaries
        r"^.*{}\W*v?\W*{}$".format(package, version_formatted_for_regex),
    ]

    for pattern in patterns:
        tags = list(filter(lambda tag: re.compile(pattern).match(tag.name.strip().lower()), tags))
        if len(tags) == 1:
            return tags[0].commit


def get_go_module_path(package):
    """assumption: package name starts with <host>/org/repo"""
    return "/".join(package.split("/")[3:])


def get_version_diff_stats_from_repository_tags(package, repo_url, old, new):
    output = VersionDifferOutput()
    output.old_version = old
    output.new_version = new

    url = sanitize_repo_url(repo_url)
    with tempfile.TemporaryDirectory() as temp_dir:
        repo = Repo.clone_from(url, temp_dir)
        tags = repo.tags

        old_commit = get_commit_of_release(tags, package, old)
        new_commit = get_commit_of_release(tags, package, new)

        if old_commit and new_commit:

            output.old_version_git_sha = old_commit
            output.new_version_git_sha = new_commit
            output.diff = get_diff_stats(temp_dir, old_commit, new_commit)

            output.new_version_filelist = get_repository_file_list(temp_dir, new_commit)
            output.old_version_filelist = get_repository_file_list(temp_dir, old_commit)

    return output


def filter_go_package_files(package, files):
    module_path = get_go_module_path(package)
    if module_path:
        files = {k: v for (k, v) in files.items() if k.startswith(module_path)}
    return files


def filter_nuget_package_files(package, files):
    package = package.lower()
    subpath = None
    for file in files.keys():
        file = file.lower().split("/")
        if package in file:
            subpath = package
            break
        else:
            for path in file:
                if package.endswith(path):
                    temp = package[: -len(path)]
                    if temp[-1] == ".":
                        subpath = path
                        break
    if subpath:
        files = dict(filter(lambda x: subpath in x[0].lower().split("/"), files.items()))

    return files


def get_version_diff_stats(ecosystem, package, old, new, repo_url=None):
    if ecosystem == GO:
        assert repo_url, "Repository URL required for Go packages"
        output = get_version_diff_stats_from_repository_tags(package, repo_url, old, new)
        output.diff = filter_go_package_files(package, output.diff)
    elif ecosystem == NUGET:
        assert repo_url, "Repository URL required for NuGet packages"
        output = get_version_diff_stats_from_repository_tags(package, repo_url, old, new)
        output.diff = filter_nuget_package_files(package, output.diff)
    else:
        output = get_version_diff_stats_registry(ecosystem, package, old, new)

    return output


def get_git_sha_from_cargo_crate(package_path):
    try:
        with open(join(package_path, ".cargo_vcs_info.json"), "r") as f:
            data = json.load(f)
            return data["git"]["sha1"]
    except:
        return None


def get_version_diff_stats_registry(ecosystem, package, old, new):
    output = VersionDifferOutput()
    output.old_version = old
    output.new_version = new

    with tempfile.TemporaryDirectory() as temp_dir_old, tempfile.TemporaryDirectory() as temp_dir_new:
        url = get_package_version_source_url(ecosystem, package, old)
        if url:
            old_path = download_package_source(url, ecosystem, package, old, temp_dir_old)
            # currently only cargo provides git sha
            if ecosystem == CARGO:
                output.old_version_git_sha = get_git_sha_from_cargo_crate(old_path)
        else:
            return output

        url = get_package_version_source_url(ecosystem, package, new)
        if url:
            new_path = download_package_source(url, ecosystem, package, new, temp_dir_new)
            # currently only cargo provides git sha
            if ecosystem == CARGO:
                output.new_version_git_sha = get_git_sha_from_cargo_crate(new_path)
        else:
            return output

        repo_old, oid_old = init_git_repo(old_path)
        repo_new, oid_new = init_git_repo(new_path)

        setup_remote(repo_old, new_path)

        output.diff = get_diff_stats(old_path, oid_old, oid_new)

        output.new_version_filelist = get_repository_file_list(new_path, oid_new)
        output.old_version_filelist = get_repository_file_list(old_path, oid_old)

    return output


def init_git_repo(path):
    repo = init_repository(path)
    index = repo.index
    index.add_all()
    tree = index.write_tree()
    sig1 = Signature("user", "email@domain.com", int(time()), 0)
    oid = repo.create_commit("refs/heads/master", sig1, sig1, "Initial commit", tree, [])
    return repo, oid


def setup_remote(repo, url):
    remote_name = "remote"
    repo.create_remote(remote_name, url)
    remote = repo.remotes[remote_name]
    remote.connect()
    remote.fetch()


def process_patch_filepath(filepath):
    filepath = filepath.removeprefix("a/")
    filepath = filepath.removeprefix("b/")
    if filepath == "/dev/null":
        filepath = None
    return filepath


def get_diff_stats_from_git_diff(uni_diff_text):
    patch_set = PatchSet(uni_diff_text)

    files = {}

    for patched_file in patch_set:
        file_path = patched_file.path  # file name

        ad_lines = [
            line.value for hunk in patched_file for line in hunk if line.is_added and line.value.strip() != ""
        ]  # the row number of deleted lines
        lines_added = len(ad_lines)

        del_lines = [
            line.value for hunk in patched_file for line in hunk if line.is_removed and line.value.strip() != ""
        ]  # the row number of added liens
        lines_removed = len(del_lines)

        if lines_added + lines_removed > 0:
            files[file_path] = FileDiff(
                process_patch_filepath(patched_file.source_file),
                process_patch_filepath(patched_file.target_file),
                patched_file.is_rename,
                lines_added,
                lines_removed,
                ad_lines,
                del_lines,
            )

    return files


def get_diff_stats(repo_path, commit_a, commit_b):
    repository = Repo(repo_path)

    uni_diff_text = repository.git.diff(str(commit_a), str(commit_b), ignore_blank_lines=True, ignore_space_at_eol=True)

    return get_diff_stats_from_git_diff(uni_diff_text)


def get_repository_file_list(repo_path, commit):
    repo = Repo(repo_path)
    head = repo.head.object.hexsha

    repo.git.checkout(commit, force=True)
    filelist = []
    for root, dirs, files in os.walk(repo_path):
        for file in files:
            if not relpath(root, repo_path).startswith(".git"):
                filelist.append(relpath(join(root, file), repo_path))

    repo.git.checkout(head, force=True)
    return set(filelist)



CARGO = "Cargo"
COMPOSER = "Composer"
GO = "Go"
MAVEN = "Maven"
NPM = "npm"
NUGET = "NuGet"
PYPI = PIP = "pip"
RUBYGEMS = "RubyGems"
ecosystems = [CARGO, COMPOSER, GO, MAVEN, NPM, NUGET, PYPI, RUBYGEMS]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['version_differ']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.24,<4.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'click>=8.0.0,<9.0.0',
 'pygit2>=1.7.2,<2.0.0',
 'requests>=2.26.0,<3.0.0',
 'rich>=10.3.0,<11.0.0',
 'unidiff>=0.7.0,<0.8.0',
 'urllib3>=1.26.7,<2.0.0']

entry_points = \
{'console_scripts': ['version-differ = version_differ.__main__:main']}

setup_kwargs = {
    'name': 'version-differ',
    'version': '0.3.5',
    'description': 'Accurate diffing between two versions of a project',
    'long_description': 'version-differ\n===========================\n\n|PyPI| |Python Version| |License| |Read the Docs| |Build| |Tests| |Codecov| |pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/version-differ.svg\n   :target: https://pypi.org/project/version-differ/\n   :alt: PyPI\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/version-differ\n   :target: https://pypi.org/project/version-differ\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/github/license/nasifimtiazohi/version-differ\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/version-differ/latest.svg?label=Read%20the%20Docs\n   :target: https://version-differ.readthedocs.io/\n   :alt: Read the documentation at https://version-differ.readthedocs.io/\n.. |Build| image:: https://github.com/nasifimtiazohi/version-differ/workflows/Build%20version-differ%20Package/badge.svg\n   :target: https://github.com/nasifimtiazohi/version-differ/actions?workflow=Package\n   :alt: Build Package Status\n.. |Tests| image:: https://github.com/nasifimtiazohi/version-differ/workflows/Run%20version-differ%20Tests/badge.svg\n   :target: https://github.com/nasifimtiazohi/version-differ/actions?workflow=Tests\n   :alt: Run Tests Status\n.. |Codecov| image:: https://codecov.io/gh/nasifimtiazohi/version-differ/branch/master/graph/badge.svg\n   :target: https://codecov.io/gh/nasifimtiazohi/version-differ\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\n\nFeatures\n--------\n\n* Given any two versions of a package, returns the list of changed files with the count of loc_added and loc_removed in each file.\n\n* Covers eight ecosystems, namely Cargo, Composer, Go, Maven, npm, NuGet, pip, and RubyGems.\n\n* For Cargo, Composer, Maven, npm, pip, and RubyGems, version-differ downloads source code for a version of a package directly from the respective package registries to measure the diff.\n\n* For Go and NuGet, it clones the source code repository, applies some heuristics to detect package specific files, and measures the diff.\n\n* diffing is performed using native git-diff, ignores black lines (does not ignore comments).\n\n\n\nInstallation\n------------\n\nYou can install *version-differ* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install version-differ\n\n\nUsage\n-----\n\nPlease see the `Command-line Reference <Usage_>`_ for details.\n\n\nCredits\n-------\n\nThis package was created with cookietemple_ using Cookiecutter_ based on Hypermodern_Python_Cookiecutter_.\n\n.. _cookietemple: https://cookietemple.com\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _PyPI: https://pypi.org/\n.. _Hypermodern_Python_Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _pip: https://pip.pypa.io/\n.. _Usage: https://version-differ.readthedocs.io/en/latest/usage.html\n',
    'author': 'Nasif Imtiaz',
    'author_email': 'nasifimtiaz88@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nasifimtiazohi/version-differ',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)
